package lcsb.mapviewer.converter.model.celldesigner.species;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import lcsb.mapviewer.common.Pair;
import lcsb.mapviewer.common.XmlParser;
import lcsb.mapviewer.common.exception.InvalidXmlSchemaException;
import lcsb.mapviewer.converter.model.celldesigner.CellDesignerElementCollection;
import lcsb.mapviewer.converter.model.celldesigner.structure.CellDesignerElement;
import lcsb.mapviewer.converter.model.celldesigner.structure.CellDesignerRna;
import lcsb.mapviewer.converter.model.celldesigner.structure.fields.CellDesignerModificationResidue;
import lcsb.mapviewer.converter.model.celldesigner.structure.fields.ModificationType;
import lcsb.mapviewer.model.map.species.Rna;
import lcsb.mapviewer.model.map.species.field.ModificationResidue;

/**
 * Class that performs parsing of the CellDesigner xml for
 * {@link CellDesignerRna} object.
 * 
 * @author Piotr Gawron
 * 
 */

public class RnaXmlParser extends AbstractElementXmlParser<CellDesignerRna, Rna> {

  /**
   * Collection of {@link CellDesignerElement cell designer elements} parsed from
   * xml.
   */
  private CellDesignerElementCollection elements;

  private ModificationResidueXmlParser modificationResidueXmlParser;

  /**
   * Default constructor. Model is required because some nodes require access to
   * other parts of the model.
   * 
   * @param elements
   *          collection of {@link CellDesignerElement cell designer elements}
   *          parsed from xml
   */
  public RnaXmlParser(final CellDesignerElementCollection elements) {
    this.elements = elements;
    this.modificationResidueXmlParser = new ModificationResidueXmlParser(elements);
  }

  @Override
  public Pair<String, CellDesignerRna> parseXmlElement(final Node rnaNode) throws InvalidXmlSchemaException {
    CellDesignerRna rna = new CellDesignerRna();
    String identifier = XmlParser.getNodeAttr("id", rnaNode);
    rna.setName(decodeName(XmlParser.getNodeAttr("name", rnaNode)));
    NodeList list = rnaNode.getChildNodes();
    for (int i = 0; i < list.getLength(); i++) {
      Node node = list.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE) {
        if (node.getNodeName().equals("celldesigner:notes")) {
          rna.setNotes(getRap().getNotes(node));
        } else if (node.getNodeName().equals("celldesigner:listOfRegions")) {
          NodeList residueList = node.getChildNodes();
          for (int j = 0; j < residueList.getLength(); j++) {
            Node residueNode = residueList.item(j);
            if (residueNode.getNodeType() == Node.ELEMENT_NODE) {
              if (residueNode.getNodeName().equalsIgnoreCase("celldesigner:region")) {
                rna.addRegion(getRnaRegion(residueNode));
              } else {
                throw new InvalidXmlSchemaException(
                    "Unknown element of celldesigner:listOfRegions " + residueNode.getNodeName());
              }
            }
          }
        } else {
          throw new InvalidXmlSchemaException("Unknown element of celldesigner:rna " + node.getNodeName());
        }
      }
    }
    return new Pair<String, CellDesignerRna>(identifier, rna);
  }

  @Override
  public String toXml(final Rna rna) {
    String attributes = "";
    String result = "";
    attributes += " id=\"r_" + elements.getElementId(rna) + "\"";
    if (!rna.getName().equals("")) {
      attributes += " name=\"" + XmlParser.escapeXml(encodeName(rna.getName())) + "\"";
    }
    result += "<celldesigner:RNA" + attributes + ">";

    if (rna.getModificationResidues().size() > 0) {
      result += "<celldesigner:listOfRegions>";
      for (final ModificationResidue mr : rna.getModificationResidues()) {
        result += modificationResidueXmlParser.toXml(mr);
      }
      result += "</celldesigner:listOfRegions>";
    }

    result += "</celldesigner:RNA>";
    return result;
  }

  /**
   * Parses CellDesigner xml node for RnaRegion.
   *
   * @param residueNode
   *          xml node to parse
   * @return {@link CellDesignerModificationResidue} object created from the node
   * @throws InvalidXmlSchemaException
   *           thrown when input xml node doesn't follow defined schema
   */
  private CellDesignerModificationResidue getRnaRegion(final Node residueNode) throws InvalidXmlSchemaException {
    CellDesignerModificationResidue residue = new CellDesignerModificationResidue();
    residue.setIdModificationResidue(XmlParser.getNodeAttr("id", residueNode));
    residue.setSize(XmlParser.getNodeAttr("size", residueNode));
    residue.setPos(XmlParser.getNodeAttr("pos", residueNode));
    String typeString = XmlParser.getNodeAttr("type", residueNode);
    if (typeString != null) {
      residue.setModificationType(ModificationType.getByCellDesignerName(typeString));
    }
    NodeList list = residueNode.getChildNodes();
    for (int i = 0; i < list.getLength(); i++) {
      Node node = list.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE) {
        throw new InvalidXmlSchemaException("Unknown element of celldesigner:region " + node.getNodeName());
      }
    }
    return residue;
  }

}
