package lcsb.mapviewer.services.impl;

import java.sql.SQLException;
import java.util.Calendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.ThreadFactory;

import javax.annotation.PostConstruct;

import org.apache.commons.lang3.exception.ExceptionUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.SpringApplication;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.jdbc.datasource.init.ScriptException;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.annotation.cache.CacheQueryMinervaJob;
import lcsb.mapviewer.annotation.cache.GeneralCacheInterface;
import lcsb.mapviewer.common.MinervaConfigurationHolder;
import lcsb.mapviewer.model.cache.CacheQuery;
import lcsb.mapviewer.model.cache.CacheType;
import lcsb.mapviewer.model.job.MinervaJob;
import lcsb.mapviewer.model.job.MinervaJobExecutor;
import lcsb.mapviewer.model.job.MinervaJobPriority;
import lcsb.mapviewer.model.job.MinervaJobStatus;
import lcsb.mapviewer.model.job.MinervaJobType;
import lcsb.mapviewer.persist.CustomDatabasePopulator;
import lcsb.mapviewer.persist.dao.MinervaJobDao;
import lcsb.mapviewer.persist.dao.cache.CacheTypeDao;
import lcsb.mapviewer.services.interfaces.IMinervaJobService;

@Service
public class MinervaJobService implements IMinervaJobService, ApplicationContextAware {

  private ApplicationContext context;

  private Logger logger = LogManager.getLogger();

  private MinervaJobDao minervaJobDao;

  private CacheTypeDao cacheTypeDao;

  private GeneralCacheInterface cache;

  private ExecutorService jobService;

  private ExecutorService jobScheduler = Executors.newSingleThreadExecutor();

  @Autowired
  private IMinervaJobService self;

  private Map<MinervaJobType, MinervaJobExecutor> executors = new HashMap<>();

  private static final Set<Integer> jobsInExecutor = new HashSet<>();

  private CustomDatabasePopulator flyway;

  private MinervaConfigurationHolder configurationHolder;

  @Autowired
  public MinervaJobService(final MinervaJobDao minervaJobDao, final GeneralCacheInterface cache, final CacheTypeDao cacheTypeDao,
      final CustomDatabasePopulator flyway, final MinervaConfigurationHolder configurationHolder) {
    this.minervaJobDao = minervaJobDao;
    this.cache = cache;
    this.cacheTypeDao = cacheTypeDao;
    this.flyway = flyway;
    this.configurationHolder = configurationHolder;
  }

  @PostConstruct
  public void init() {

    try {
      flyway.populate(flyway.getDataSource().getConnection());
    } catch (ScriptException | SQLException e) {
      logger.error(e, e);
    }

    jobService = Executors.newScheduledThreadPool(configurationHolder.getBatchThreads(), new ThreadFactory() {
      @Override
      public Thread newThread(final Runnable r) {
        Thread t = new Thread(r);
        t.setDaemon(true);
        return t;
      }
    });
    jobScheduler.execute(() -> {
      int errorIssues = 0;
      while (true) {
        try {
          synchronized (jobsInExecutor) {
            if (((ScheduledThreadPoolExecutor) jobService).getQueue().size() == 0) {
              if (self.getPending(1).size() == 0) {
                addMissingRefreshTasks(100);
              }
              for (MinervaJob job : self.getPending(10)) {
                if (!jobsInExecutor.contains(job.getId())) {
                  jobsInExecutor.add(job.getId());
                  self.evict(job);
                  jobService.submit(createJobExecutionWrapper(job));
                }
              }
            }
          }
          Thread.sleep(100);
          errorIssues = 0;
        } catch (InterruptedException e) {
          logger.error("QUEUE interrupted", e);
          break;
        } catch (Throwable e) {
          logger.error("Unexpected exception", e);
          errorIssues++;
          if (errorIssues == 5) {
            break;
          }
          try {
            // wait for 30 seconds - we don't know what is happening probably something with
            // db
            Thread.sleep(30000);
          } catch (InterruptedException interruption) {
            logger.error("QUEUE interrupted", e);
            break;
          }
          logger.info("Recovering attempt: " + errorIssues);

        }
      }
      logger.fatal("QUEUE stopped. Stopping minerva");
      int exitCode = SpringApplication.exit(context, () -> 1);
      System.exit(exitCode);
    });

    registerExecutor(MinervaJobType.REFRESH_CACHE, cache);
  }

  @Override
  public void setApplicationContext(final ApplicationContext ctx) throws BeansException {
    this.context = ctx;

  }

  @Override
  @Transactional
  public void addJob(final MinervaJob job) {
    minervaJobDao.add(job);
  }

  private Callable<Object> createJobExecutionWrapper(final MinervaJob job) {
    return new Callable<Object>() {
      @Override
      public Object call() {
        try {
          startJob(job);

          MinervaJobExecutor executor = getExecutor(job.getJobType());
          if (executor == null) {
            logger.error("Cannot find executor for job: " + job.getJobType() + " (job id: " + job.getId() + ")");
            job.setJobStatus(MinervaJobStatus.FAILED);
            job.setLogs("Cannot find executor for job: " + job.getJobType() + " (job id: " + job.getId() + ")");
            return self.updateJob(job);
          }
          executor.execute(job);

          return finishJob(job);
        } catch (Throwable e) {
          return finishJobWithProblems(job, e);
        } finally {
          synchronized (jobsInExecutor) {
            jobsInExecutor.remove(job.getId());
          }
        }
      }

    };
  }

  protected MinervaJobExecutor getExecutor(final MinervaJobType jobType) {
    return executors.get(jobType);
  }

  @Override
  public void registerExecutor(final MinervaJobType jobType, final MinervaJobExecutor executor) {
    executors.put(jobType, executor);
  }

  private Calendar lastRefreshTaskExceptionTimestamp = null;

  private void addMissingRefreshTasks(final int limit) {
    try {
      List<CacheQuery> queries = cache.getExpired(limit);
      for (CacheQuery cacheQuery : queries) {
        CacheQueryMinervaJob params = new CacheQueryMinervaJob(cacheQuery.getQuery(), cacheQuery.getType());
        MinervaJob job = new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.LOWEST, params);
        self.addJob(job);
      }
    } catch (Exception e) {
      // log these errors not too often (once per 10 minutes)
      Calendar tenMinutesAgo = Calendar.getInstance();
      tenMinutesAgo.add(Calendar.MINUTE, -10);
      if (lastRefreshTaskExceptionTimestamp == null || lastRefreshTaskExceptionTimestamp.before(tenMinutesAgo)) {
        logger.error("Sever problem with adding tasks to refresh list", e);
        lastRefreshTaskExceptionTimestamp = Calendar.getInstance();
      }
    }
  }

  private void startJob(final MinervaJob job) {
    logger.debug("Starting job: " + job.getId() + ", " + job.getJobType() + ", priority: " + job.getPriority() + ", " + job.getJobParameters());
    job.setJobStatus(MinervaJobStatus.RUNNING);
    self.updateJob(job);
  }

  private MinervaJob finishJob(final MinervaJob job) {
    logger.debug("Job finished: " + job.getId());
    job.setJobStatus(MinervaJobStatus.DONE);
    return self.updateJob(job);
  }

  private MinervaJob finishJobWithProblems(final MinervaJob job, final Throwable e) {
    logger.error("Problem with executing job: " + job.getJobType() + " (job id: " + job.getId() + ")", e);
    job.setJobStatus(MinervaJobStatus.FAILED);
    job.setLogs(ExceptionUtils.getStackTrace(e));
    return self.updateJob(job);
  }

  @Override
  @Transactional
  public MinervaJob updateJob(final MinervaJob job) {
    minervaJobDao.saveOrUpdate(job);
    return job;
  }

  @Override
  public void waitForTasksToFinish() {
    synchronized (jobsInExecutor) {
      if (((ScheduledThreadPoolExecutor) jobService).getQueue().size() == 0) {
        if (self.getPending(1).size() == 0) {
          addMissingRefreshTasks(Integer.MAX_VALUE);
        }
      }
    }

    logger.debug("Waiting for jobs to finish");
    while (self.getCount(MinervaJobStatus.PENDING) + self.getCount(MinervaJobStatus.RUNNING) > 0) {
      logger.debug("Waiting...");
      try {
        Thread.sleep(100);
      } catch (InterruptedException e) {
        logger.error("Problem with waiting", e);
        return;
      }
    }
    logger.debug("Waiting finished");
  }

  @Override
  @Transactional
  public List<MinervaJob> getPending() {
    return minervaJobDao.getPending();
  }

  @Override
  @Transactional
  public List<MinervaJob> getPending(final int max) {
    return minervaJobDao.getPending(max);
  }

  @Override
  @Transactional
  public CacheType getTypeById(final Integer typeId) {
    return cacheTypeDao.getById(typeId);
  }

  @Override
  @Transactional
  public void evict(final MinervaJob job) {
    minervaJobDao.evict(job);
  }

  @Override
  @Transactional
  public long getCount(final MinervaJobStatus status) {
    return minervaJobDao.getCount(status);
  }

  @Override
  @Transactional
  public void setInterruptedOnRunningJobs() {
    for (MinervaJob job : minervaJobDao.getRunning()) {
      job.setJobStatus(MinervaJobStatus.INTERRUPTED);
      minervaJobDao.update(job);
    }

  }

  @Override
  @Transactional
  public List<MinervaJob> getAllJobs() {
    return minervaJobDao.getAll();
  }

  @Override
  @Transactional
  public void removeOldJobs(final Calendar beforeDate) {
    minervaJobDao.deleteBefore(MinervaJobStatus.PENDING, beforeDate);

  }

}
