package lcsb.mapviewer.services.impl;

import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.util.Calendar;
import java.util.List;

import org.junit.After;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;

import lcsb.mapviewer.annotation.cache.CacheQueryMinervaJob;
import lcsb.mapviewer.annotation.cache.GeneralCacheInterface;
import lcsb.mapviewer.annotation.services.annotators.HgncAnnotator;
import lcsb.mapviewer.common.tests.TestUtils;
import lcsb.mapviewer.common.tests.UnitTestFailedWatcher;
import lcsb.mapviewer.model.cache.CacheQuery;
import lcsb.mapviewer.model.cache.CacheType;
import lcsb.mapviewer.model.job.MinervaJob;
import lcsb.mapviewer.model.job.MinervaJobParameters;
import lcsb.mapviewer.model.job.MinervaJobPriority;
import lcsb.mapviewer.model.job.MinervaJobType;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.services.SpringServiceTestConfig;
import lcsb.mapviewer.services.interfaces.IMinervaJobService;
import lcsb.mapviewer.services.interfaces.IMiriamService;
import lcsb.mapviewer.services.jobs.RefreshMiriamInfoMinervaJob;

@ContextConfiguration(classes = SpringServiceTestConfig.class)
@RunWith(SpringJUnit4ClassRunner.class)
public class MinervaJobServiceTest extends TestUtils {

  @Autowired
  private IMinervaJobService minervaJobService;

  @Autowired
  private IMiriamService miriamService;

  @Autowired
  private GeneralCacheInterface cache;

  @Autowired
  private HgncAnnotator hgncAnnotator;

  private String query = "https://google.com/";

  private CacheType type;

  @Rule
  public UnitTestFailedWatcher unitTestFailedWatcher = new UnitTestFailedWatcher();

  @Before
  public final void setUp() throws Exception {
    type = hgncAnnotator.getCacheType();
  }

  @After
  public final void tearDown() throws Exception {
    minervaJobService.waitForTasksToFinish();
    cache.clearCache();
  }

  @Test
  public void testRefreshTask() throws Exception {
    CacheQuery entry = cache.getByQuery(query, type);
    if (entry != null) {
      cache.delete(entry);
    }

    MinervaJobParameters params = new CacheQueryMinervaJob(query, type.getId());

    minervaJobService.addJob(new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.MEDIUM, params));
    minervaJobService.addJob(new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.MEDIUM, params));
    minervaJobService.addJob(new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.MEDIUM, params));
    minervaJobService.addJob(new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.LOWEST, params));
    minervaJobService.addJob(new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.HIGH, params));

    minervaJobService.waitForTasksToFinish();

    entry = cache.getByQuery(query, type);

    assertNotNull(entry);

    assertTrue(Calendar.getInstance().before(entry.getExpires()));
    List<MinervaJob> jobs = minervaJobService.getAllJobs();
    MinervaJob low = null;
    MinervaJob high = null;
    for (MinervaJob minervaJob : jobs) {
      if (minervaJob.getPriority() == MinervaJobPriority.HIGH.getValue()) {
        high = minervaJob;
      }
      if (minervaJob.getPriority() == MinervaJobPriority.LOWEST.getValue()) {
        low = minervaJob;
      }
    }
    assertTrue(high.getJobStarted().before(low.getJobStarted()));
  }

  @Test
  public void testFailedRefreshTaskRemoveFromCache() throws Exception {
    String invalidQuery = "invalidQuery";
    cache.setCachedQuery(invalidQuery, type, "some val");

    MinervaJobParameters params = new CacheQueryMinervaJob(invalidQuery, type.getId());

    assertNotNull(cache.getByQuery(invalidQuery, type));

    minervaJobService.addJob(new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.MEDIUM, params));

    minervaJobService.waitForTasksToFinish();

    assertNull(cache.getByQuery(query, type));

  }

  @Test
  public void testCreateRefreshTaskForExpiredCache() throws Exception {
    cache.setCachedQuery(query, type, "invalidated");
    cache.invalidateByQuery(query, type);

    minervaJobService.waitForTasksToFinish();

    CacheQuery entry = cache.getByQuery(query, type);

    assertNotEquals(entry.getValue(), "invalidated");
  }

  @Test
  public void testRefreshMiriamInfoTask() throws Exception {
    CacheQuery entry = cache.getByQuery(query, type);
    if (entry != null) {
      cache.delete(entry);
    }

    MiriamData md = new MiriamData(MiriamType.PUBMED, "12345");
    miriamService.add(md);

    RefreshMiriamInfoMinervaJob params = new RefreshMiriamInfoMinervaJob(md.getId());
    MinervaJob job = new MinervaJob(MinervaJobType.REFRESH_MIRIAM_INFO, MinervaJobPriority.MEDIUM, params);

    minervaJobService.addJob(job);
    minervaJobService.waitForTasksToFinish();

    md = miriamService.getById(md.getId());
    assertNotNull(md.getArticle());
    assertNotNull(md.getLink());
  }
}
