package lcsb.mapviewer.services.impl;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.CALLS_REAL_METHODS;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Set;

import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.core.impl.Log4jLogEvent;
import org.apache.logging.log4j.message.SimpleMessage;
import org.apache.poi.util.IOUtils;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.annotation.Rollback;

import lcsb.mapviewer.commands.CopyCommand;
import lcsb.mapviewer.common.MinervaLoggerAppender;
import lcsb.mapviewer.converter.ComplexZipConverter;
import lcsb.mapviewer.converter.ComplexZipConverterParams;
import lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser;
import lcsb.mapviewer.converter.zip.ModelZipEntryFile;
import lcsb.mapviewer.converter.zip.ZipEntryFile;
import lcsb.mapviewer.model.LogMarker;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.ProjectLogEntry;
import lcsb.mapviewer.model.ProjectLogEntryType;
import lcsb.mapviewer.model.cache.UploadedFileEntry;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.model.SubmodelType;
import lcsb.mapviewer.model.map.species.GenericProtein;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.persist.dao.MinervaJobDao;
import lcsb.mapviewer.persist.dao.map.ModelDao;
import lcsb.mapviewer.services.ServiceTestFunctions;
import lcsb.mapviewer.services.interfaces.IFileService;
import lcsb.mapviewer.services.interfaces.IMinervaJobService;
import lcsb.mapviewer.services.interfaces.IModelService;
import lcsb.mapviewer.services.interfaces.IUserService;
import lcsb.mapviewer.services.utils.CreateProjectParams;

@Rollback(true)
public class ProjectServiceTest extends ServiceTestFunctions {

  private final String tmpResultDir = "tmp/";

  @Autowired
  protected IModelService modelService;

  @Autowired
  protected IUserService userService;

  @Autowired
  protected IFileService fileService;

  @Autowired
  protected ModelDao modelDao;

  @Autowired
  protected MinervaJobDao minervaJobDao;

  @Autowired
  protected IMinervaJobService minervaJobService;

  @Before
  public void setUp() throws Exception {
  }

  @After
  public void tearDown() throws Exception {
    minervaJobService.waitForTasksToFinish();
    if (projectService.getProjectByProjectId(TEST_PROJECT_ID) != null) {
      projectService.removeProject(TEST_PROJECT_ID, tmpResultDir);
    }
    minervaJobDao.clearTable();
  }

  @Test
  public void testCopyComplexWithCompartments() throws Exception {
    ZipEntryFile entry1 = new ModelZipEntryFile("main.xml", "main", true, false, SubmodelType.UNKNOWN);
    ZipEntryFile entry2 = new ModelZipEntryFile("s1.xml", "s1", false, false, SubmodelType.UNKNOWN);
    ZipEntryFile entry3 = new ModelZipEntryFile("s2.xml", "s2", false, false, SubmodelType.UNKNOWN);
    ZipEntryFile entry4 = new ModelZipEntryFile("s3.xml", "s3", false, false, SubmodelType.UNKNOWN);
    ZipEntryFile entry5 = new ModelZipEntryFile("mapping.xml", "mapping", false, true, SubmodelType.UNKNOWN);

    CreateProjectParams params = createProjectParams("testFiles/complexModel/complex_model_with_compartment.zip")
        .addZipEntry(entry1).addZipEntry(entry2).addZipEntry(entry3).addZipEntry(entry4).addZipEntry(entry5)
        .analyzeAnnotations(false);

    ComplexZipConverter parser = new ComplexZipConverter(CellDesignerXmlParser.class);
    ComplexZipConverterParams complexParams = new ComplexZipConverterParams().zipFile(fileService.getById(params.getProjectFileId()));
    for (final ZipEntryFile entry : params.getZipEntries()) {
      complexParams.entry(entry);
    }
    Model model = parser.createModel(complexParams);

    assertNotNull(model);
    assertEquals("main", model.getName());
    assertEquals(3, model.getSubmodelConnections().size());

    int oldSize = 0;
    for (final Model m : model.getSubmodels()) {
      oldSize += m.getParentModels().size();
    }

    new CopyCommand(model).execute();

    int newSize = 0;
    for (final Model m : model.getSubmodels()) {
      newSize += m.getParentModels().size();
    }
    assertEquals("Submodels doesn't match after copying", oldSize, newSize);
  }

  private CreateProjectParams createProjectParams(final String filename) throws FileNotFoundException, IOException {
    InputStream is = new FileInputStream(filename);
    return createProjectParams(filename, is);
  }

  private CreateProjectParams createProjectParams(final String filename, final InputStream is) throws IOException {
    UploadedFileEntry fileEntry = new UploadedFileEntry();
    fileEntry.setLocalPath(filename);
    byte[] data = IOUtils.toByteArray(is);
    fileEntry.setLength(data.length);
    fileEntry.setFileContent(data);
    fileService.add(fileEntry);

    return new CreateProjectParams()
        .setUser(userService.getUserByLogin(ADMIN_BUILT_IN_LOGIN))
        .projectId(TEST_PROJECT_ID)
        .version("")
        .complex(filename.endsWith("zip"))
        .projectFile(fileEntry.getId())
        .annotations(true)
        .images(false)
        .projectDir(tmpResultDir)
        .analyzeAnnotations(true)
        .parser(new CellDesignerXmlParser());
  }

  @Test
  public void testCreateLogEntries() throws Exception {
    Protein protein = new GenericProtein("id");
    MinervaLoggerAppender appender = MinervaLoggerAppender.createAppender();
    appender.append(Log4jLogEvent.newBuilder()
        .setLevel(Level.WARN)
        .setMessage(new SimpleMessage("Hello world"))
        .setMarker(new LogMarker(ProjectLogEntryType.OTHER, protein))
        .build());
    appender.append(Log4jLogEvent.newBuilder()
        .setLevel(Level.ERROR)
        .setMessage(new SimpleMessage("Hello world 2"))
        .setMarker(new LogMarker(ProjectLogEntryType.OTHER, protein))
        .build());
    appender.append(Log4jLogEvent.newBuilder()
        .setLevel(Level.ERROR)
        .setMessage(new SimpleMessage("Hello world 3"))
        .build());

    ProjectService service = Mockito.mock(ProjectService.class, Mockito.CALLS_REAL_METHODS);
    Set<ProjectLogEntry> entries = service.createLogEntries(appender);
    assertEquals(3, entries.size());
  }

  @Test
  public void testCreateIgnoredLogEntries() throws Exception {
    MinervaLoggerAppender appender = MinervaLoggerAppender.createAppender();
    appender.append(Log4jLogEvent.newBuilder()
        .setLevel(Level.WARN)
        .setMessage(new SimpleMessage("Hello world"))
        .setMarker(LogMarker.IGNORE)
        .build());
    ProjectService service = Mockito.mock(ProjectService.class, Mockito.CALLS_REAL_METHODS);
    Set<ProjectLogEntry> entries = service.createLogEntries(appender);
    assertEquals(0, entries.size());
  }

  @Test
  public void testFixProjectIssuesWithElementWithoutModel() throws Exception {
    ProjectService service = Mockito.mock(ProjectService.class, CALLS_REAL_METHODS);
    Project project = new Project();
    ModelData model = new ModelData();
    GenericProtein protein = new GenericProtein("1");
    project.addModel(model);
    model.addElement(protein);
    service.fixProjectIssues(project);
    assertTrue(project.getLogEntries().size() > 0);
  }

}
