package lcsb.mapviewer.web.app;

import java.awt.Color;
import java.io.File;
import java.util.Calendar;
import java.util.List;
import java.util.concurrent.TimeUnit;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.context.annotation.Configuration;
import org.springframework.http.CacheControl;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.json.MappingJackson2HttpMessageConverter;
import org.springframework.scheduling.annotation.EnableScheduling;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.ViewResolver;
import org.springframework.web.servlet.config.annotation.CorsRegistry;
import org.springframework.web.servlet.config.annotation.InterceptorRegistry;
import org.springframework.web.servlet.config.annotation.ResourceHandlerRegistry;
import org.springframework.web.servlet.config.annotation.ViewControllerRegistry;
import org.springframework.web.servlet.config.annotation.WebMvcConfigurer;
import org.springframework.web.servlet.view.InternalResourceViewResolver;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.module.SimpleModule;

import lcsb.mapviewer.common.MinervaConfigurationHolder;
import lcsb.mapviewer.modelutils.serializer.CalendarSerializer;
import lcsb.mapviewer.modelutils.serializer.ColorSerializer;

@Configuration
@EnableScheduling
@ComponentScan(basePackages = { "lcsb.mapviewer.api" })
public class WebConfig implements WebMvcConfigurer {

  private Logger logger = LogManager.getLogger();

  @Autowired
  private MinervaConfigurationHolder configurationHolder;

  @Override
  public void addCorsMappings(final CorsRegistry registry) {
    registry.addMapping("/**").allowedMethods("*");
  }

  @Override
  public void addResourceHandlers(final ResourceHandlerRegistry registry) {
    String path = getDataRootPath();
    CacheControl cacheControl = CacheControl.maxAge(31, TimeUnit.DAYS);
    registry.addResourceHandler("/**")
        .addResourceLocations("/", "classpath:/")
        .setCacheControl(cacheControl);

    registry.addResourceHandler("/minerva/**")
        .addResourceLocations("/", "classpath:/")
        .setCacheControl(cacheControl);

    registry.addResourceHandler("/map_images/**")
        .addResourceLocations("file:" + path + "/map_images/")
        .setCacheControl(cacheControl);

    registry.addResourceHandler("/minerva-big/**")
        .addResourceLocations("file:" + path + "/minerva-big/")
        .setCacheControl(cacheControl);
  }

  private String getDataRootPath() {
    String path = configurationHolder.getDataPath();
    if (path == null) {
      path = new File(lcsb.mapviewer.common.Configuration.getJarPath()).getParent();
    }
    logger.debug("Data path: " + path);
    return path;
  }

  private List<HandlerInterceptor> interceptors;

  @Autowired
  public WebConfig(final List<HandlerInterceptor> interceptors) {
    this.interceptors = interceptors;
  }

  @Override
  public void extendMessageConverters(final List<HttpMessageConverter<?>> converters) {
    for (HttpMessageConverter<?> converter : converters) {
      if (converter instanceof MappingJackson2HttpMessageConverter) {
        MappingJackson2HttpMessageConverter m = (MappingJackson2HttpMessageConverter) converter;
        SimpleModule module = new SimpleModule();
        module.addSerializer(Calendar.class, new CalendarSerializer());
        module.addSerializer(Color.class, new ColorSerializer());

        m.getObjectMapper().registerModule(module);
      }
    }
  }

  @Bean
  public ViewResolver viewResolver() {
    InternalResourceViewResolver resolver = new InternalResourceViewResolver();
    resolver.setPrefix("/");
    resolver.setSuffix(".html");
    return resolver;
  }

  @Override
  public void addInterceptors(final InterceptorRegistry registry) {
    interceptors.forEach(registry::addInterceptor);
  }

  @Override
  public void addViewControllers(final ViewControllerRegistry registry) {
    registry
        .addViewController("/")
        .setViewName("index");
  }

  @Bean
  public ObjectMapper objectMapper() {
    ObjectMapper mapper = new ObjectMapper();

    return mapper;
  }

}
